module simplecpu
	#(parameter addrSize = 10)
	(input clk,
	 input reset,
	 inout [7:0] dataBus,
	 output [addrSize-1:0] addrBus,
	 output reg we);

// non-register control/internal signals
reg incAddr;
reg [addrSize-1:0] addrsaveNext, addrNext;
wire [addrSize-1:0] addr;
reg [7:0] aNext;
reg [7:0] opcodeNext;
reg aluflagZNext;
reg aluflagCNext;
reg aluflagVNext;
reg aluflagNNext; 
reg [7:0] dataOut;

// internal CPU registers
counter #(.width(addrSize)) u0(.clk(clk), .reset(reset), .enable(incAddr), .d(addrNext), .q(addr));
reg [addrSize-1:0] addrsave;
reg [7:0] opcode;
reg [7:0] a;
reg aluflagZ;
reg aluflagC;
reg aluflagV;
reg aluflagN; 
wire [3:0] aluflags = { aluflagZ, aluflagC, aluflagV, aluflagN };
wire [2:0] mathop = opcode[7:5];
wire [1:0] addrmode = {opcode[4], opcode[2]}; 
					
// states
reg [3:0] state, stateNext;
localparam [3:0] fetchOp = 4'b0000,
				 skipByte0 = 4'b0100,
				 skipByte1 = 4'b0101,
				 setFlags = 4'b0110,
				 doMath = 4'b0111,
				 fetchAddress0 = 4'b1000,
				 fetchAddress1 = 4'b1001;		

// addressing modes
localparam [1:0] immediate = 2'b00,
			     absolute = 2'b01,
			     indirectY = 2'b10,
			     absoluteX = 2'b11;
			     
// math ops
localparam [2:0] opORA = 3'b000,
				 opAND = 3'b001,
				 opEOR = 3'b010,
				 opADC = 3'b011,
				 opSTA = 3'b100,
				 opLDA = 3'b101,
				 opCMP = 3'b110,
				 opSBC = 3'b111;

// ALU
wire [7:0] aluResult;
wire aluCarryOut;
wire aluAddSub = ~mathop[2];
wire aluCarryIn = aluflagC | ~mathop[0]; // force the cin during CMP instruction
alu u1(.a(a), .b(dataBus), .cin(aluCarryIn), .addSub(aluAddSub), .result(aluResult), .cout(aluCarryOut));
				 
always @* begin
	// by default, registers should maintain their current state
	addrsaveNext = addrsave;
	addrNext = addr;
	incAddr = 1'b0;
	stateNext = state;
	aluflagZNext = aluflagZ;
	aluflagCNext = aluflagC;
	aluflagVNext = aluflagV;
	aluflagNNext = aluflagN;
	opcodeNext = opcode;
	aNext = a;
	
	// default combinatorial output values 
	dataOut = 8'bzzzzzzzz;
	we = 1'b1; 
	
	case (state)
		fetchOp: 
			begin
				opcodeNext = dataBus;
				incAddr = 1'b1;
				
				casez (dataBus) 
					8'b???10000: 
						// conditional branches have the form xxy10000, 
						// where xx is the flag and y is the value to match
						if (aluflags[dataBus[7:6]] == dataBus[5]) 
							stateNext = fetchAddress0; // branch taken
						else
							stateNext = skipByte0; // branch not taken
					8'b??????01: 
						// logic and mathematical ops using the accumulator have the form aaabxb01,
						// where aaa is the operation and bb is the addressing mode
						// aaa: 0=ORA, AND, EOR, ADC, STA, LDA, CMP, 7=SBC
						// bb: 0=immediate, absolute, (ZP),Y, 3=absolute,X 
						begin
							if ({dataBus[4], dataBus[2]} == 2'b00)
								// immediate addressing: go directly to math state
								stateNext = doMath;
							else
								// other address mode: setup the address first
								stateNext = fetchAddress0;
						end
					8'h4C: // absolute jump
						stateNext = fetchAddress0;
					8'hB8: // set ALU flags
						stateNext = setFlags;
					default: // treat unrecognized oaddrsaveodes as HALT
						begin
							incAddr = 1'b0;
							stateNext = fetchOp;
						end
				endcase
			end
		skipByte0: 
			begin
				incAddr = 1'b1;
				stateNext = skipByte1;
			end
		skipByte1:
			begin
				incAddr = 1'b1;
				stateNext = fetchOp;
			end
		setFlags:
			begin
				incAddr = 1'b1;
				{ aluflagZNext, aluflagCNext, aluflagVNext, aluflagNNext } = (aluflags | dataBus[3:0]) & dataBus[7:4];
				stateNext = fetchOp;
			end
		fetchAddress0:
			begin
				incAddr = 1'b1;
				addrsaveNext[7:0] = dataBus;
				stateNext = fetchAddress1;
			end
		fetchAddress1:
			begin
				addrsaveNext = addr;
				addrNext[7:0] = addrsave[7:0];
				addrNext[addrSize-1:8] = dataBus[addrSize-9:0];
				
				// Can reach this state during a jmp/branch or during a math operation.
				// All jmp/branch instructions end in 00, math ops ends in 01 or 10.
				if (opcode[0] | opcode[1]) 
					stateNext = doMath;
				else
					stateNext = fetchOp;
			end	
		doMath:
			begin
				case (mathop)
					opORA:
						begin
							aNext = a | dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opAND:
						begin
							aNext = a & dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opEOR:
						begin
							aNext = a ^ dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opLDA:
						begin
							aNext = dataBus;
							aluflagNNext = aNext[7];
							aluflagZNext = (aNext == 0);
						end
					opSTA: 
						begin
							we = 0;
							dataOut = a;
						end
					opADC:
						begin
							aNext = aluResult;
							aluflagCNext = aluCarryOut;
							aluflagNNext = aluResult[7];
							aluflagZNext = (aluResult == 0);
							aluflagVNext = (~a[7] & ~dataBus[7] & aluResult[7]) |
										   (a[7] & dataBus[7] & ~aluResult[7]);					
						end
					opSBC: 
						begin
							aNext = aluResult;
							aluflagCNext = aluCarryOut;
							aluflagNNext = aluResult[7];
							aluflagZNext = (aluResult == 0);
							aluflagVNext = (~a[7] & dataBus[7] & aNext[7]) |
										   (a[7] & ~dataBus[7] & ~aNext[7]);
						end	
					opCMP:
						begin
							aluflagCNext = aluCarryOut;
							aluflagNNext = aluResult[7];
							aluflagZNext = (aluResult == 0);
						end	
				endcase
				
				// Can reach this state during a memory reference, or during immediate mode.
				// Memory reference needs to restore the saved program counter.
				if (addrmode != immediate) begin
					addrNext = addrsave;
					stateNext = skipByte1;
				end else begin
					incAddr = 1'b1;
					stateNext = fetchOp;
				end
			end
		default:
			stateNext = fetchOp;	
	endcase
end

always @(posedge clk or negedge reset) begin
	if (!reset) begin
		state <= fetchOp;
	end
	else begin
		state <= stateNext;
		addrsave <= addrsaveNext;				
		aluflagZ <= aluflagZNext;
		aluflagC <= aluflagCNext;
		aluflagV <= aluflagVNext;
		aluflagN <= aluflagNNext;
		a <= aNext;
		opcode <= opcodeNext;
	end
end

assign dataBus = dataOut;
assign addrBus = addr;

endmodule
